local cql = ClassicQuestLog

local questList = {} -- ordered list of QuestLogInfo results for headers and quests
_questList = questList
local numQuests = 0 -- count of questIDs displayed in the list
local somethingExpanded = false -- true if any quest header is expanded

function cql.Log:Setup()
    self:SetScript("OnSizeChanged",self.OnSizeChanged)
    self:OnSizeChanged(self:GetWidth(),self:GetHeight())
end

-- adds the given questID to cql.settings.QuestHistory
function cql.Log:AddQuestToHistory(questID)
    local questHistory = cql.settings.QuestHistory
    if questID and questID>0 then
        -- remove any prior instances of this being selected
        for i=#questHistory,1,-1 do
            if questHistory[i]==questID then
                tremove(questHistory,i)
            end
        end
        -- and add this questID to the end
        tinsert(questHistory,questID)
    end
end

function cql.Log:Update()
    local content = self.ScrollFrame.Content
    local questHistory = cql.settings.QuestHistory

    self.ScrollFrame.AllButton:SetNormalAtlas(somethingExpanded and "Campaign_HeaderIcon_Open" or "Campaign_HeaderIcon_Closed")
    self.ScrollFrame.AllButton:SetPushedAtlas(somethingExpanded and "Campaign_HeaderIcon_OpenPressed" or "Campaign_HeaderIcon_ClosedPressed")
    self.ScrollFrame.AllButton:SetEnabled(numQuests>0)

    if cql.settings.HighContrast then
        self.Back:SetAtlas("QuestDetailsBackgrounds-Accessibility_Dark")
        self.Back:SetVertexColor(0.5,0.5,0.5)
    else
        self.Back:SetAtlas("QuestLogBackground")
        self.Back:SetVertexColor(1,1,1)
    end

    -- if this is an update and there are no quests, look for some
    if #questList==0 then
        self:UpdateQuestList()
    end

    local questID = self:GetSelectedQuest()

    -- if this is different than last quest selected, add to history
    if questID~=0 and (#questHistory==0 or questID~=questHistory[#questHistory]) then
        cql.Log:AddQuestToHistory(questID)
    end

    -- if selected quest wasn't found, then pick the recent one if any
    if not self:IsQuestInLog(questID) and self:GetNumQuests()>0 then
        local nextQuestID,nextHeader = self:GetRecentQuestID()
        -- if next quest is in a header that's collapsed, then expand the header
        if nextHeader and cql.settings.CollapsedHeaders[nextHeader] then
            cql.settings.CollapsedHeaders[nextHeader] = nil
            self:UpdateQuestList()
        end
        if nextQuestID and nextQuestID>0 then
            C_QuestLog.SetSelectedQuest(nextQuestID)
        end
    end

    -- place the buttons
    local contentWidth = content.width
    for index,info in ipairs(questList) do
        if not content.Buttons[index] then
            content.Buttons[index] = CreateFrame("Button",nil,content,"ClassicQuestLogListButtonTemplate")
            content.Buttons[index]:SetPoint("TOPLEFT",content.Buttons[index-1],"BOTTOMLEFT")
            content.Buttons[index]:SetWidth(contentWidth)
        end
        local button = content.Buttons[index]
        button.index = index
        button.info = info
        button:SetText(info.name or info.title)
        if info.isHeader then
            button:FillHeader(info)
            button:Show()
        elseif info.questID then
            button:FillNormal(info)
            button:Show()
        else
            button:Hide()
        end
    end

    -- hide any unused buttons
    for index=#questList+1,#content.Buttons do
        content.Buttons[index]:Hide()
        content.Buttons[index].info = nil
    end

end

function cql.Log:OnSizeChanged(width,height)
    local content = self.ScrollFrame.Content
    local contentWidth = max(floor(width+0.5)-35,150)
    content:SetWidth(contentWidth)
    content.width = contentWidth

    for _,button in ipairs(content.Buttons) do
        button:SetWidth(contentWidth)
    end

    self:Update()
end

function cql.Log.ScrollFrame.AllButton:OnClick()
    if somethingExpanded then
        for _,info in ipairs(questList) do
            if info.isHeader then
                cql.settings.CollapsedHeaders[info.title] = true
            end
        end
    else
        wipe(cql.settings.CollapsedHeaders)
    end
    cql.Log:UpdateQuestList()
    cql.Log:Update()
    cql.Dialog:Hide()
end

function cql.Log:IsSomethingExpanded()
    return somethingExpanded
end

function cql.Log:GetNumQuests()
    return numQuests
end

-- returns the currently selected questID, or 0 if no quest selected
-- use this instead of C_QuestLog.GetSelectedQuest() so it can swap the selected
-- quest to 0 if it's not known (when the quest log is empty,
-- C_QuestLog.GetSelectedQuest() can otherwise return a questID not in the log)
function cql.Log:GetSelectedQuest()
    local questID = C_QuestLog.GetSelectedQuest()
    if questID and questID>0 and (C_QuestLog.IsOnQuest(questID) or C_QuestLog.IsQuestDisabledForSession(questID)) then
        return questID
    else
        return 0
    end
end

-- removes the last occurring header from questList that has the given title
function cql.Log:RemoveHeader(title)
    for i=#questList,1,-1 do
        if questList[i].isHeader and questList[i].title==title then
            tremove(questList,i)
            return
        end
    end
end

-- updates questList with headers and non-hidden quests in an expanded header
function cql.Log:UpdateQuestList()
    wipe(questList)
    numQuests = 0
    somethingExpanded = false

    local currentHeader = nil
    local headerHasQuests = false
    for i=1,C_QuestLog.GetNumQuestLogEntries() do
        local info = C_QuestLog.GetInfo(i)
        if info.isHeader and not info.isHidden then
            -- if we are encountering a header (after first) and there's no quest in prior header, remove prior header
            if currentHeader and not headerHasQuests then
                self:RemoveHeader(currentHeader)
            end
            -- add this new header to questList
            tinsert(questList,info)
            currentHeader = info.title
            headerHasQuests = false
        elseif not info.isHidden then
            if info.questID and info.questID > 0 then
                headerHasQuests = true
                numQuests = numQuests + 1
                if not currentHeader or not cql.settings.CollapsedHeaders[currentHeader] then
                    tinsert(questList,info)
                    somethingExpanded = true
                end
            end
        end
    end
    -- if last header has only hidden quests, remove it
    if currentHeader and not headerHasQuests then
        self:RemoveHeader(currentHeader)
    end
end


-- turns on/off tracking for given questID
function cql.Log:ToggleQuestWatch(questID)
    if not questID then
        questID = self:GetSelectedQuest()
    end
    if questID > 0 then
        if C_QuestLog.GetQuestWatchType(questID) then
            C_QuestLog.RemoveQuestWatch(questID)
        else
            C_QuestLog.AddQuestWatch(questID)
        end
    end
end

-- returns true if the questID is in the log and not hidden
function cql.Log:IsQuestInLog(questID)
    local index = C_QuestLog.GetLogIndexForQuestID(questID)
    if index then
        local info = C_QuestLog.GetInfo(index)
        if not info.isHidden then
            return true
        end
    end
    return false
end

-- returns the header for the questID
function cql.Log:GetQuestHeader(questID)
    if self:IsQuestInLog(questID) then
        local index = C_QuestLog.GetLogIndexForQuestID(questID)
        for i=index-1,1,-1 do
            local info = C_QuestLog.GetInfo(i)
            if info.isHeader then
                return info.title
            end
        end
    end
end

-- finds the most recently selected questID in history (and its header) that's still valid
function cql.Log:GetRecentQuestID()
    local questHistory = cql.settings.QuestHistory
    for i=#questHistory,1,-1 do
        local questID = questHistory[i]
        if questID > 0 and numQuests > 0 and self:IsQuestInLog(questID) and (C_QuestLog.IsOnQuest(questID) or C_QuestLog.IsQuestDisabledForSession(questID)) then
            local lastHeader
            for i=1,C_QuestLog.GetNumQuestLogEntries() do
                local info = C_QuestLog.GetInfo(i)
                if info.isHeader then
                    lastHeader = info.title
                elseif info.questID==questID and not info.isHidden then
                    return info.questID,lastHeader
                end
            end
        end
    end
    -- if we reached here, then we couldn't find one in history that's valid, return first in log
    return self:GetNextQuestID(0)
end

-- for picking the next quest in the log after one is about to be abandoned, or first
-- questID in log if questID is 0; returns questID,header title,true if it was bottommost quest
function cql.Log:GetNextQuestID(questID)
    if not questID then
        questID = cql.Log:GetSelectedQuest()
    end
    local nextQuestID, lastHeader, sawQuestID
    for i=1,C_QuestLog.GetNumQuestLogEntries() do
        local info = C_QuestLog.GetInfo(i)
        if info.isHeader then
            lastHeader = info.title
        elseif info.questID and info.questID > 0 and not info.isHidden then
            if questID==0 then -- if picking next from 0, then return first unhidden one found
                return info.questID,lastHeader,false
            elseif info.questID==questID then -- if found matching questID, set flag that it's seen
                sawQuestID = true
            else
                nextQuestID = info.questID
                if sawQuestID then
                    return nextQuestID,lastHeader,false
                end
            end
        end
    end
    -- if reached here, then questID was last non-hidden quest in the list
    if nextQuestID then
        return nextQuestID,lastHeader,true
    end
end

-- if quest accepted is a visible one in the log (not an auto-accept world quest etc)
-- then switch to that quest
function cql.Log:QuestAccepted(questID)
    local lastHeader
    for i=1,C_QuestLog.GetNumQuestLogEntries() do
        local info = C_QuestLog.GetInfo(i)
        if info.isHeader then
            lastHeader = info.title
        elseif info.questID and not info.isHidden and info.questID==questID then
            if lastHeader then
                cql.settings.CollapsedHeaders[lastHeader] = nil
            end
            C_QuestLog.SetSelectedQuest(questID)
            if cql:IsVisible() then
                cql.timer:Start(0.1,cql.DelayedUpdateAndScrollToSelectedQuest)
            end
        end
    end
end

-- returns true if the given index is fully within the visible scrollframe
function cql.Log:IsIndexInView(index)
    if index then
        local scrollPercent = self.ScrollFrame.ScrollBar:GetScrollPercentage()
        local buttonHeight = floor(self.ScrollFrame.Content.Buttons[1]:GetHeight()+0.5)
        local dataHeight = buttonHeight * #questList
        local scrollHeight = floor(self.ScrollFrame:GetHeight()+0.5)
        local topOffset = scrollPercent*max(0,dataHeight-scrollHeight) -- if not scrollable, this is always 0
        local bottomOffset = topOffset + scrollHeight
        local indexTop = buttonHeight * (index-1)
        local indexBottom = buttonHeight * (index)
        return indexTop >= topOffset and indexBottom <= bottomOffset
    end
end

-- scroll questList index into view (used by ScrollToQuestID)
function cql.Log:ScrollToIndex(index)
    local buttonHeight = floor(self.ScrollFrame.Content.Buttons[1]:GetHeight()+0.5)
    local dataHeight = buttonHeight * #questList
    local scrollHeight = floor(self.ScrollFrame:GetHeight()+0.5)
    if index and dataHeight > scrollHeight then
        local indexTop = buttonHeight * (index-1)
        local indexBottom = buttonHeight * (index)
        local topOffset = indexBottom - (scrollHeight/2) + (indexBottom-indexTop)/2
        local scrollPercent = min(1,max(0,topOffset/(dataHeight-scrollHeight)))
        self.ScrollFrame.ScrollBar:SetScrollPercentage(scrollPercent)
    end
end

-- returns the index in questList that the quest is in
function cql.Log:GetQuestIDIndex(questID)
    for index,info in ipairs(questList) do
        if info.questID==questID then
            return index
        end
    end
end

-- scrolls the log to bring the given questID into view
function cql.Log:ScrollToQuestID(questID)
    if questID and questID~=0 then
        local header = self:GetQuestHeader(questID)
        -- if questID is in a collapsed header, expand the header
        if header and cql.settings.CollapsedHeaders[header] then
            cql.settings.CollapsedHeaders[header] = nil
            self:UpdateQuestList()
            self:Update()
        end
        -- find the index for the questID (should now be in expanded header) and scroll to it
        local index = self:GetQuestIDIndex(questID)
        if index and not self:IsIndexInView(index) then
            self:ScrollToIndex(index)
        end
    end
end

-- to be called from a timer, scrolls the log to the currently-selected questID
function cql.Log:DelayedScrollToSelectedQuestID()
    cql.Log:ScrollToQuestID(cql.Log:GetSelectedQuest())
end

